@@ -14,6 +14,7 @@ import tempfile
 import toolchest_client
 
 from .containers import DockerContainer
+from .module_detection import get_modules_to_register
 
 
 def patch_system_call(user_docker_container_name=None, original_function=None, pass_kwargs=True,
@@ -51,21 +52,31 @@ def patch_system_call(user_docker_container_name=None, original_function=None, p
     return run
 
 
-def find_and_replace_function(member, unique_docstring, replace_with, depth=0, unpatch=False):
-    if depth > 3:
+def find_and_replace_function(member, unique_docstring, replace_with, patched_functions, depth=0, max_depth=3):
+    if depth > max_depth:
         return
     if type(member) in [str, type(None)]:
         return
     elif type(member) == dict:
-        for name, sub_attribute in member.items():
+        for name, sub_attribute in member.copy().items():  # The dict can change size, so make a copy for iteration
             if name == "__builtins__":
                 continue
-            this_one = find_and_replace_function(sub_attribute, unique_docstring, replace_with, depth + 1, unpatch)
+            this_one = find_and_replace_function(
+                member=sub_attribute,
+                unique_docstring=unique_docstring,
+                replace_with=replace_with,
+                patched_functions=patched_functions,
+                depth=depth + 1,
+                max_depth=max_depth,
+            )
             if this_one:
-                if unpatch:
-                    member[name] = member[name].lug_original_function
-                else:
-                    member[name] = replace_with
+                patched_functions.append({
+                    "member": member,
+                    "attribute_name": name,
+                    "original_value": replace_with.lug_original_function,
+                    "is_dictionary": True,
+                })
+                member[name] = replace_with
                 return False
     elif isinstance(member, types.ModuleType):
         # import subprocess -> subprocess is a module
@@ -77,25 +88,39 @@ def find_and_replace_function(member, unique_docstring, replace_with, depth=0, u
         for name, sub_member in members:
             if name == "__builtins__":
                 continue
-            this_one = find_and_replace_function(sub_member, unique_docstring, replace_with, depth + 1, unpatch)
+            this_one = find_and_replace_function(
+                member=sub_member,
+                unique_docstring=unique_docstring,
+                replace_with=replace_with,
+                patched_functions=patched_functions,
+                depth=depth + 1,
+                max_depth=max_depth,
+            )
             if this_one:
-                if unpatch:
-                    patched_function = getattr(member, name)
-                    setattr(member, name, patched_function.lug_original_function)
-                else:
-                    setattr(member, name, replace_with)
+                patched_functions.append({
+                    "member": member,
+                    "attribute_name": name,
+                    "original_value": getattr(member, name),
+                    "is_dictionary": False,
+                })
+                setattr(member, name, replace_with)
                 return False
         return False
     else:
-        # Base case; catches:
-        # os.system (a "builtin_function_or_method")
-        # subprocess.run (a "function")
         try:
-            if unpatch:
-                if hasattr(member, "lug_original_function"):
-                    return True
-            elif hasattr(member, "__doc__") and unique_docstring in member.__doc__:
+            if hasattr(member, "__doc__") and member.__doc__ and unique_docstring in member.__doc__:
                 return True
+            elif isinstance(member, types.FunctionType):
+                # Explore the globals of any imported function
+                find_and_replace_function(
+                    member=member.__globals__,
+                    unique_docstring=unique_docstring,
+                    replace_with=replace_with,
+                    patched_functions=patched_functions,
+                    depth=depth + 1,
+                    max_depth=5,  # If we're exploring a function globals, increase max depth to five
+                )
+                return False
         except (TypeError, ModuleNotFoundError):
             return False
         return False
@@ -124,48 +149,65 @@ def patch_system_calls(func, user_docker_container_name, docker_shell_location):
         docker_shell_location=docker_shell_location,
     )
 
+    patched_functions = []
     find_and_replace_function(
         member=func.__globals__,
         unique_docstring="Execute a child program in a new process.",
-        replace_with=new_subprocess_popen
+        replace_with=new_subprocess_popen,
+        patched_functions=patched_functions,
     )
     if not hasattr(subprocess.Popen, "is_lug_function"):
         find_and_replace_function(
             member=func.__globals__,
             unique_docstring="Run command with arguments and return a CompletedProcess instance",
-            replace_with=new_subprocess_run
+            replace_with=new_subprocess_run,
+            patched_functions=patched_functions,
         )
     find_and_replace_function(
         member=func.__globals__,
         unique_docstring="Execute the command in a subshell",
-        replace_with=new_os_system
+        replace_with=new_os_system,
+        patched_functions=patched_functions,
     )
 
-
-def unpatch_system_calls(func):
-    find_and_replace_function(
-        member=func.__globals__,
-        unique_docstring=None,
-        replace_with=None,
-        unpatch=True
-    )
+    return patched_functions
 
 
-def create_python_script(func, args, kwargs, temp_input, user_docker, docker_shell_location):
+def unpatch_system_calls(patched_functions):
+    for patched_function_info in patched_functions:
+        if patched_function_info["is_dictionary"]:
+            attribute = patched_function_info["attribute_name"]
+            patched_function_info["member"][attribute] = patched_function_info["original_value"]
+        else:
+            setattr(
+                patched_function_info["member"],
+                patched_function_info["attribute_name"],
+                patched_function_info["original_value"],
+            )
+
+
+def create_python_script(func, args, kwargs, temp_input, user_docker, docker_shell_location, serialize_dependencies):
     output_uuid = uuid.uuid4()
     with open(temp_input.name, 'w') as fp:
-        cloudpickle.register_pickle_by_value(inspect.getmodule(func))
-        cloudpickle.register_pickle_by_value(sys.modules[__name__])
-        pickled_patch = cloudpickle.dumps(patch_system_calls)
-        encoded_pickled_patch = base64.encodebytes(pickled_patch)
+        if serialize_dependencies:
+            modules_to_register = get_modules_to_register(func)
+            for module in modules_to_register:
+                cloudpickle.register_pickle_by_value(module)
+        else:
+            cloudpickle.register_pickle_by_value(sys.modules[__name__])
         pickled_func = cloudpickle.dumps(func)
         encoded_func = base64.encodebytes(pickled_func)
+        pickled_patch = cloudpickle.dumps(patch_system_calls)
+        encoded_pickled_patch = base64.encodebytes(pickled_patch)
         pickled_args = cloudpickle.dumps(args)
         encoded_args = base64.encodebytes(pickled_args)
         pickled_kwargs = cloudpickle.dumps(kwargs)
         encoded_kwargs = base64.encodebytes(pickled_kwargs)
-        cloudpickle.unregister_pickle_by_value(inspect.getmodule(func))
-        cloudpickle.unregister_pickle_by_value(sys.modules[__name__])
+        if serialize_dependencies:
+            for module in modules_to_register:
+                cloudpickle.unregister_pickle_by_value(module)
+        else:
+            cloudpickle.unregister_pickle_by_value(sys.modules[__name__])
         fp.write("import base64\n")
         fp.write("import cloudpickle\n")
         fp.write(f"func = cloudpickle.loads(base64.decodebytes({encoded_func}))\n")
@@ -194,19 +236,28 @@ def parse_toolchest_run(output_path, output_uuid):
 
 
 def execute_remote(func, args, kwargs, toolchest_key, remote_output_directory, tmp_dir, image, remote_inputs,
-                   user_docker, remote_instance_type, volume_size, python_version, docker_shell_location):
+                   user_docker, remote_instance_type, volume_size, python_version, docker_shell_location,
+                   serialize_dependencies):
     if not os.path.exists(tmp_dir):
         os.mkdir(tmp_dir)
     temp_input = tempfile.NamedTemporaryFile(dir=tmp_dir)
     temp_directory = None
-    output_uuid = create_python_script(func, args, kwargs, temp_input, user_docker, docker_shell_location)
+    output_uuid = create_python_script(
+        func=func,
+        args=args,
+        kwargs=kwargs,
+        temp_input=temp_input,
+        user_docker=user_docker,
+        docker_shell_location=docker_shell_location,
+        serialize_dependencies=serialize_dependencies,
+    )
     try:
         if toolchest_key:
             toolchest_client.set_key(toolchest_key)
         if remote_output_directory is None:
             temp_directory = tempfile.TemporaryDirectory(dir=tmp_dir)
         output_directory = remote_output_directory or temp_directory.name
-        run = toolchest_client.lug(
+        remote_run = toolchest_client.lug(
             custom_docker_image_id=image,
             tool_version=python_version,
             inputs=remote_inputs,
@@ -216,7 +267,7 @@ def execute_remote(func, args, kwargs, toolchest_key, remote_output_directory, t
             instance_type=remote_instance_type,
             volume_size=volume_size,
         )
-        status_response = run.get_status(return_error=True)
+        status_response = remote_run.get_status(return_error=True)
         status = status_response['status']
         if status == 'failed':
             raise ValueError(f"Remote execution failed due to: {status_response['error_message']}")
@@ -246,14 +297,17 @@ def execute_local(mount, client, user_docker, func, args, kwargs, docker_shell_l
         signal.signal(signal.SIGHUP, user_docker.signal_kill_handler)
         signal.signal(signal.SIGSEGV, user_docker.signal_kill_handler)
         signal.signal(signal.SIGTERM, user_docker.signal_kill_handler)
-    patch_system_calls(func, user_docker.container_name, docker_shell_location)
-    result = func(*args, **kwargs)
+    patched_functions = patch_system_calls(func, user_docker.container_name, docker_shell_location)
+    try:
+        result = func(*args, **kwargs)
+    finally:
+        unpatch_system_calls(patched_functions)
     return result
 
 
 def run(image, mount=os.getcwd(), tmp_dir=os.getcwd(), docker_shell_location="/bin/sh", remote=False,
         remote_inputs=None, remote_output_directory=None, toolchest_key=None, remote_instance_type=None,
-        volume_size=None):
+        volume_size=None, serialize_dependencies=False):
     def decorator_lug(func):
         @functools.wraps(func)
         def inner(*args, **kwargs):
@@ -271,27 +325,26 @@ def run(image, mount=os.getcwd(), tmp_dir=os.getcwd(), docker_shell_location="/b
             supported_versions = ["3.7", "3.8", "3.9", "3.10", "3.11"]
             if python_version not in supported_versions:
                 raise ValueError(f"Python version {python_version} is not supported. PRs welcome!")
-
             try:
                 if remote:
-                    result = execute_remote(func=func, args=args, kwargs=kwargs,
-                                            image=image, remote_inputs=remote_inputs, toolchest_key=toolchest_key,
-                                            remote_output_directory=remote_output_directory, tmp_dir=tmp_dir,
-                                            user_docker=user_docker, remote_instance_type=remote_instance_type,
-                                            volume_size=volume_size, python_version=python_version,
-                                            docker_shell_location=docker_shell_location)
+                    result = execute_remote(
+                        func=func, args=args, kwargs=kwargs, image=image, remote_inputs=remote_inputs,
+                        toolchest_key=toolchest_key, remote_output_directory=remote_output_directory, tmp_dir=tmp_dir,
+                        user_docker=user_docker, remote_instance_type=remote_instance_type, volume_size=volume_size,
+                        python_version=python_version, docker_shell_location=docker_shell_location,
+                        serialize_dependencies=serialize_dependencies
+                    )
                 else:
-                    result = execute_local(func=func, args=args, kwargs=kwargs,
-                                           mount=mount, client=client, user_docker=user_docker,
-                                           docker_shell_location=docker_shell_location)
+                    result = execute_local(
+                        func=func, args=args, kwargs=kwargs, mount=mount, client=client, user_docker=user_docker,
+                        docker_shell_location=docker_shell_location
+                    )
             finally:
                 if user_docker is not None and user_docker.container is not None:
                     user_docker.container.reload()
                     if user_docker.container.status == "running":
                         user_docker.container.stop()
                     user_docker.container.remove()
-                if not remote:
-                    unpatch_system_calls(func)
             return result
         return inner
     return decorator_lug
